#ifndef NET_MINE_H
#define NET_MINE_H

#include <stdlib.h> /* for size_t only */

/**
 * Net-MineSeeper protocol specifiction.
 *
 * $Id: net-mine.h,v 1.4 2005/05/16 10:39:21 tea Exp tea $
 */


/**
 * Basic data type. instruction code, numbers of rows/cols, and flags.
 * are all 1 byte.
 */
typedef unsigned char net_mine_t;

/**
 * The size which stores rows and cols are an unsigned char type,
 * MAX_COLS, MAX_ROWS must lesser than 2**8-1
 */
#define MAX_COLS    255
#define MAX_ROWS    255

/**
 * Instructions set. stored in "type" field of net-mine packet.
 *
 * * FLAG type has a parameter.
 *   flag on ... 0x11, unflag 0x11
 * 
 * * NUMS type has a parameter the below 4 bits.
 *   For example, when the number is 3, the type is 0x83.
 *   that is used as the mask for the bits.
 */
#define NOOP        0x00
#define INIT        0x01
#define QUIT        0x02
#define FLAG        0x10
#define OPEN        0x20
#define BOMB        0x40
#define NUMS        0x80
#define BLANK       (NUMS+0)

/**
 * Check pointer of arguments, and return NULL if it was wrong pointer.
 */
#define BOARD_CHECK(_board) \
    if ((_board) == NULL) return NULL

#define PACKET_CHECK(_packet) \
    if ((_packet) == NULL) return NULL

/**
 * Accessor macro for the Board item.
 */
#define BOARD_ITEM(_board,_col,_row) \
    ((_board)->screen[(_col) * (_board)->rows + (_row)])

#define BOARD_ITEM_WITH_NUM(_board,_col,_row) \
    (BOARD_ITEM((_board),(_col),(_row)) | \
     (net_mine_board_count_bomb((_board),(_col),(_row)) & NUMS))

/** MineBoard#screen loop convention.
 * 
 *  BOARD_LOOP_INIT_VARS(column, row);
 *  BOARD_LOOP_ITER(board, column, row) {
 *      BOARD_ITEM(board, column, row) = 0;
 *  }
 * 
 *  BOARD_LOOP_INIT_VARS(index, board_size);
 *  BOARD_LOOP_WITH_INDEX(board, index, board_size) {
 *      board->screen[index] = 0;
 *  }
 * 
 */
#define BOARD_LOOP_INIT_VARS(_var1,_var2) \
    size_t _var1## = 0, _var2## = 0
        
#define BOARD_LOOP_COLS(_board,_var) \
    for (##_var = 0; ##_var < (_board)->cols; ++##_var)
        
#define BOARD_LOOP_ROWS(_board,_var) \
    for (##_var = 0; ##_var < (_board)->rows; ++##_var)
                
#define BOARD_LOOP_ITER(_board,_colvar,_rowvar) \
    BOARD_LOOP_ROWS((_board),##_rowvar) BOARD_LOOP_COLS((_board),##_colvar)

#define BOARD_LOOP_WITH_INDEX(_board,_var,_sizevar) \
    _sizevar## = (_board)->cols * (_board)->rows; \
    for (##_var = 0; ##_var < ##_sizevar; ++##var)
    

#define PACKET_SET(_packet,_board,_col,_row) \
    { (_packet)->type = BOARD_ITEM_WITH_NUM((_board),(_col),(_row)); \
      (_packet)->col  = (_col); \
      (_packet)->row  = (_row); }

#define PACKET_CLEAR(_packet) \
    { (_packet)->type = \
      (_packet)->col  = \
      (_packet)->row  = (net_mine_t) NULL; }

/**
 * condition macro
 *
 * Thise macro checks conditions of tile. as the macro name express.
 */
#define ON_BOARD(_board,_col,_row) \
    ((0 <= (_col) && (_col) < (_board)->cols) && \
     (0 <= (_row) && (_row) < (_board)->rows))

#define TILE_TYPE(_board,_col,_row,_mask) \
    (ON_BOARD((_board),(_col),(_row)) \
     ? (BOARD_ITEM((_board),(_col),(_row)) & (_mask)) : 0)

#define IS_OPENED(_board,_col,_row) \
    (TILE_TYPE((_board),(_col),(_row),OPEN))

#define IS_NOT_OPENED(_board, _col, _row) \
    (! IS_OPENED((_board),(_col),(_row)))
    
#define HAS_FLAG(_board,_col,_row) \
    (TILE_TYPE((_board),(_col),(_row),FLAG))

#define HAS_BOMB(_board,_col,_row) \
    (TILE_TYPE((_board),(_col),(_row),BOMB))

/**
 * define command function macro.
 * 
 * The parameters for this macro are, function name and instruction type.
 * when the command function will have row/col arguments,
 * then use DEFINE_COMMAND_EX, else DEFINE_COMMAND.
 * 
 * @see net-mine.c
 */
#define DEFINE_COMMAND(_name, _type) \
    MinePacket *##_name(MinePacket *packet) \
    { PACKET_CHECK(packet); \
      packet->type = (_type); \
      return packet; }

#define DEFINE_COMMAND_EX(_name, _type) \
    MinePacket *##_name(MinePacket *packet, size_t col, size_t row) \
    { PACKET_CHECK(packet); \
      packet->type = (_type); packet->col = col; packet->row = row; \
      return packet; }

/**
 * User data structures.
 */

typedef struct net_mine_board {
    net_mine_t  cols;
    net_mine_t  rows;
    net_mine_t  mines;
    net_mine_t  *screen;
    size_t      count;  /* How many tiles are opened. */
} MineBoard;

typedef struct net_mine_packet {
    net_mine_t  type;
    net_mine_t  col;
    net_mine_t  row;
    net_mine_t  session;
} MinePacket;

MineBoard *net_mine_board_init(MineBoard *board,
                                size_t cols, size_t rows);
MineBoard *net_mine_board_delete(MineBoard *board);

MinePacket *net_mine_board_open(MineBoard *board, MinePacket *packet,
                                size_t col, size_t row);
MinePacket *net_mine_board_flag(MineBoard *board, MinePacket *packet,
                                size_t col, size_t row);

MinePacket *net_mine_cmd_init(MinePacket *packet);
MinePacket *net_mine_cmd_quit(MinePacket *packet);
MinePacket *net_mine_cmd_open(MinePacket *packet, size_t col, size_t row);
MinePacket *net_mine_cmd_flag(MinePacket *packet, size_t col, size_t row);

#endif /* NET_MINE_H */
